{
  "type": "record",
  "Aspect": {
    "name": "schemaFieldsInferredNeighbors"
  },
  "name": "SchemaFieldsInferredNeighbors",
  "namespace": "com.linkedin.pegasus2avro.inferred",
  "fields": [
    {
      "type": {
        "type": "array",
        "items": {
          "type": "record",
          "name": "SchemaFieldInferredNeighbors",
          "namespace": "com.linkedin.pegasus2avro.inferred",
          "fields": [
            {
              "java": {
                "class": "com.linkedin.pegasus2avro.common.urn.Urn"
              },
              "type": "string",
              "name": "fieldUrn",
              "doc": "SchemaField Urn",
              "Urn": "Urn"
            },
            {
              "type": {
                "type": "array",
                "items": {
                  "type": "record",
                  "name": "InferredNeighbor",
                  "namespace": "com.linkedin.pegasus2avro.inferred",
                  "fields": [
                    {
                      "java": {
                        "class": "com.linkedin.pegasus2avro.common.urn.Urn"
                      },
                      "type": "string",
                      "name": "urn",
                      "doc": "Urn of inferred neighbor entity",
                      "Urn": "Urn"
                    },
                    {
                      "type": {
                        "type": "record",
                        "name": "InferredMetadataSource",
                        "namespace": "com.linkedin.pegasus2avro.inferred",
                        "fields": [
                          {
                            "type": "string",
                            "name": "algorithm",
                            "doc": "Algorithm used for inference"
                          },
                          {
                            "type": [
                              "null",
                              "float"
                            ],
                            "name": "score",
                            "default": null,
                            "doc": "Inference score - can be used to decide rank of inference\nFor same algorithm, higher score represents higher confidence in inference.\nThis score is not intended to be compared across multiple algorithms."
                          },
                          {
                            "type": [
                              "null",
                              {
                                "type": "map",
                                "values": "string"
                              }
                            ],
                            "name": "context",
                            "default": null,
                            "doc": "Additional details about this inference"
                          },
                          {
                            "type": [
                              "null",
                              {
                                "type": "array",
                                "items": {
                                  "type": "enum",
                                  "symbolDocs": {
                                    "DATASET_PLATFORM": "Platform of dataset",
                                    "DATASET_SCHEMA": "Schema of dataset",
                                    "ENTITY_DESCRIPTION": "Description of entity",
                                    "ENTITY_NAME": "Name of entity",
                                    "FIELD_DATA_TYPE": "Datatype of dataset field",
                                    "FIELD_DESCRIPTION": "Description of dataset field",
                                    "FIELD_NAME": "Name of dataset field",
                                    "FIELD_PARENT_DATASET": "Dataset containing the field",
                                    "LINEAGE": "Lineage of entity"
                                  },
                                  "name": "SimilarityFactor",
                                  "namespace": "com.linkedin.pegasus2avro.inferred",
                                  "symbols": [
                                    "ENTITY_NAME",
                                    "ENTITY_DESCRIPTION",
                                    "LINEAGE",
                                    "DATASET_SCHEMA",
                                    "DATASET_PLATFORM",
                                    "FIELD_NAME",
                                    "FIELD_DESCRIPTION",
                                    "FIELD_DATA_TYPE",
                                    "FIELD_PARENT_DATASET"
                                  ],
                                  "doc": "Factors related to an entity"
                                }
                              }
                            ],
                            "name": "similarityFactors",
                            "default": null,
                            "doc": "Primary factors contributing to this inference\nOrdering of the array indicates rank."
                          }
                        ],
                        "doc": "Provenace details corresponding to an inference"
                      },
                      "name": "source",
                      "doc": "Provenance details about this inference"
                    }
                  ],
                  "doc": "Records Urn of inferred neighbor entity and its provenance"
                }
              },
              "name": "inferredNeighbors",
              "doc": "Neighbors of a schema field as determined by algorithm"
            }
          ],
          "doc": "Neighbors for a particular schema field as inferred by various algorithms"
        }
      },
      "name": "schemaFieldsInferredNeighbors",
      "doc": "Neighbors of dataset's schema fields as determined by algorithm"
    }
  ],
  "doc": "Aspect for representing Neighbors of dataset's schema fields as determined by algorithm"
}